class ImageContainer:
    """
        This class is responsible for holding a reference to the most relevant image.
    """
    def __init__(self):
        self.img = None

    def set(self, img):
        self.img = img

    def get(self):
        return self.img


from scvf.constants import settings_keys


def pipeline_loop(image_container, pipeline_manager, settings, output_consumer, running):
    """
    This function creates a loop that manages all the pipelines.
    it is responsible for processing an image and sending the output to the output_consumer,
    as well as applying new settings to the pipelines if they arrive.
    :param image_container:
        a container that holds the latest image
    :param pipeline_manager:
        a ``PipelineManager`` instance that holds all the pipelines
    :param settings:
        the ``Settings`` instance for this loop
    :param output_consumer:
        the function that should receive the processed output of the pipeline
    :param running:
        a lambda that tells the pipeline_loop whether it should continue to run
    """
    while running():
        if settings.is_updated():
            pipeline_manager.set_pipeline(settings.get(settings_keys["pipeline_name"], ''))
        img = image_container.get()
        if img is not None:
            pipeline_manager.process(img)
            output_consumer(pipeline_manager.get_output())





def callback(camera_settings, pipeline_settings):
    """
        this function generates a callback should be called each time a new settings arrives.
        the ``settings_supplier`` is responsible for receiving new settings and calling this function when
        a new settings arrives. in turn, the callback updates the given setting sets with the setting.
    :param camera_settings:
        the settings for the ``camera_loop``
    :param pipeline_settings:
        the settings for the ``pipeline_loop``
    :return:
        a callback that updates the given settings
    """
    def c(key, value):
        camera_settings.update(**{key: value})
        pipeline_settings.update(**{key: value})
    return c


class PipelineManager:
    """
        This pipeline is responsible for managing all given pipelines.

        Note that you have to pass a setting that tells which pipeline should be used.
        otherwise the loop wouldn't use any pipeline.
    """
    def __init__(self, pipelines):
        self.pipelines = pipelines
        if len(pipelines.keys()) < 1:
            self.current_pipeline = ""
        else:
            self.current_pipeline = list(pipelines.keys())[0]

    def set_pipeline(self, pipeline_id):
        """
            changes the current pipeline to the pipeline with the given id
        :param pipeline_id:
            the id of the pipeline
        """
        if pipeline_id in self.pipelines.keys():
            self.current_pipeline = pipeline_id

    def process(self, img):
        """
        this function gives the current pipeline process a given image
        :param img:
            the image that should be processed
        :return:
        """
        pipeline = self.pipelines.get(self.current_pipeline, None)
        if pipeline is not None:
            pipeline.process(img)

    def get_output(self):
        """
            This function returns the current output of the current pipeline
        :return:
            the output of the pipeline
        """
        pipeline = self.pipelines.get(self.current_pipeline, None)
        if pipeline is not None:
            return pipeline.get_output()



from scvf.constants import settings_keys


def camera_loop(image_container, camera_manager, settings, running):
    """
    This function is responsible for constantly reading images from the camera buffer,
    so that the image available to the loop is as relevant as possible
    :param image_container:
        an object that is responsible for storing the most relevant image
    :param camera_manager:
        an instance of ``CameraManager`` that is responsible for managing all the cameras
        and reading images from them
    :param settings:
        the ``Settings`` object for the camera_loop
    :param running:
        a lambda that tells the pipeline_loop whether it should continue to run
    """
    while running():
        if settings.is_updated():
            camera_manager.set_exposure(settings.get(settings_keys["exposure"], ""))
            camera_manager.set_camera(settings.get(settings_keys["camera_id"], ""))

        img = camera_manager.get_image()
        if img is not None:
            image_container.set(img)
    camera_manager.release()


from threading import Thread

from scvf.loops import camera_loop, pipeline_loop, settings_callback
from scvf.util import PipelineManager, ImageContainer, CameraManager, Settings
import numpy

# This flag tells the loop wheather it should run. It is switched to True on start and False on end.
running = False


def start(pipelines, camera_port=0, output_consumer=lambda: None, settings_supplier=lambda x: None,
          width=320, height=240):
    """
        This function starts the CV loop, receiving all the needed parameters.

        Parameters:
            :param pipelines:
                a list of all the pipelines the loop should manage
            :param camera_port:
                the initial
            :param output_consumer
                this is the function that should receive the output generated by the pipeline
                and (optionally) send it to another computer.
            :param settings_supplier:
                this function should process a callback that should be called each time a new settings arrives
                (for more info look into the doc on loops.settings_callback)
            :param width
                the width of the image in pixels
            :param height
                the height of the image in pixels

    """
    global running
    img = ImageContainer()
    pipeline_settings = Settings()
    camera_settings = Settings()

    pipeline_manager = PipelineManager(pipelines)
    camera_manager = CameraManager(port=camera_port, width=width, height=height)

    pipeline_thread = Thread(target=pipeline_loop, args=(img, pipeline_manager, pipeline_settings,
                                                         output_consumer, lambda: running))
    camera_thread = Thread(target=camera_loop, args=(img, camera_manager, camera_settings, lambda: running))
    running = True
    settings_supplier(settings_callback(camera_settings, pipeline_settings))
    pipeline_thread.start()
    camera_thread.start()


def end():
    """
        this function ends the cv loop
    """
    global running
    running = False
